/*{{{  File banner*/
/*@(#)=====================================================*\
||@(#)  Project : GPMIMD ESPRIT P5404
||@(#)  Authors : Mark Debbage, Mark Hill and Sean Wykes
||@(#)            University of Southampton
||  
||@(#)    Title : Top level driver
||@(#)   System : Occam 2 C
||@(#) Filename : ParserDrv.c
||@(#)  Version : 1.30
||@(#)     Date : 1/14/94
\*@(#)====================================================*/
/*}}}*/
/*{{{  includes*/
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <ctype.h>

#include "Parser.h"
#include "Tree.h"
#include "OccamOut.h"
#include "TypeChkAg.h"
#include "Cout.h"
#include "CoutAg.h"
#include "ParUsage.h"
#include "ParUsageAg.h"
#include "Transform.h"
/*}}}*/
/*{{{  externs + defines*/
#define BoolChr(A) ((A)?'t':'f')
#define COMPILATION_FAILED -1
extern char *VersionName;
extern char *VersionDate;

#define DEFAULT_STR(f) (IS_Flag(f,DefaultModes) ? "disable" : "enable")
/*}}}*/
/*{{{  switches*/
#define d_yydebug false
#define d_ldebug false
#define d_nowarnings false

#define d_showtree false
#define d_showsyms false
#define d_showdecs false
#define d_showinfo false

#define d_dointrinsics true
#define d_doquery false
#define d_dooccout false
#define d_dostore true
#define d_docout true
#define d_dotransform true

#define d_dotypechk_ag true
#define d_doparusage_ag true
#define d_docout_ag true
#define d_lexonly false

int yydebug = d_yydebug;
int ldebug = d_ldebug;
int nowarnings = d_nowarnings;
bool showinfo = d_showinfo;

static bool showtree = d_showtree;
static bool showsyms = d_showsyms;
static bool showdecs = d_showdecs;

static bool dointrinsics = d_dointrinsics;
static bool doquery = d_doquery;
static bool dooccout = d_dooccout;
static bool dostore = d_dostore;
static bool docout = d_docout;
static bool dotransform = d_dotransform;

static bool dotypechk_ag = d_dotypechk_ag;
static bool doparusage_ag = d_doparusage_ag;
static bool docout_ag = d_docout_ag;
static bool lexonly = d_lexonly;
/*}}}*/

/*{{{  statics*/
char * OARCH = ARCH;
char * OSEARCH = "OSEARCH";
char * OLIB = "lib";
char * ODEFAULT_LIB = NULL;
char * OSPOC = "SPOC";
char * SLASH = "/";

FILE *yyf;
bool Compiler_is_LittleEndian;
static char *SpServerFile   = "spserver.lib";
static char *IntrinsicsFile = "Intrinsics.lib";

/*}}}*/

/*{{{  static void ApplyAttributeGrammars(void)*/
static void ApplyAttributeGrammars(void)
{
  if (ErrorCount>0) return;

  if (!CheckTree(TreeRoot)) { fprintf(stderr,"Fatal-ABSTRACT SYNTAX TREE is INVALID\n"); exit(-1); }

  /*{{{  if (dotypechk_ag)*/
  if (dotypechk_ag)
  {
    if (showinfo) fprintf(stderr,"Type checking\n") ;
    BeginTypeChkAg();
    TypeChkAg(TreeRoot);
    CloseTypeChkAg();
  }
  /*}}}*/

  if (ErrorCount>0) return;

  /*{{{  if (doparusage_ag)*/
  if (doparusage_ag)
  {
    if (showinfo) fprintf(stderr,"Parallel usage and alias checking\n") ;
    BeginParUsageAg();
    ParUsageAg(TreeRoot);
    CloseParUsageAg();
  }
  /*}}}*/

  if (ErrorCount>0) return;

  /*{{{  if (dotransform)*/
  if (dotransform)
  {
    if (showinfo) fprintf(stderr,"Tree transformations\n") ;
    
    Transform(TreeRoot);
  
    if (!CheckTree(TreeRoot)) { fprintf(stderr,"Fatal-ABSTRACT SYNTAX TREE is INVALID\n"); exit(-1); }
  }
  /*}}}*/

  if (ErrorCount>0) return;

  /*{{{  if (docout_ag)*/
  if (docout_ag)
  {
    if (showinfo) fprintf(stderr,"Computing tree attributes for C code generation\n") ;
  
    BeginCoutAg();
    CoutAg(TreeRoot);
    CloseCoutAg();
  }
  /*}}}*/
}
/*}}}*/

/*{{{  #ifndef CORE_DUMP*/
#ifndef CORE_DUMP
void  FatalError(int signal)
{
  Message("Internal Error (please report to dan@uk.ac.soton.ecs)",xxFatal,Attribute.Position);
}
#endif
/*}}}*/
int main (int argc,char **argv)
{
  FILE *infile=stdin;
  char *filename= "stdin";
  char *outfile = "";
  int DefaultModes;
  char *ospoc = getenv(OSPOC);

  ldebug = false;
  yyf = stdout;
  
#ifndef CORE_DUMP
  signal(SIGSEGV,FatalError);
#endif

  /*{{{  default modes*/
  {
    int val = 1;
    Compiler_is_LittleEndian = (((char *)&val)[0]) == 1;
  }
  ADD_Mode(xxOccamLines);
  ADD_Mode(xxCheckRanges);
  ADD_Mode(xxCheckArrays);
  ADD_Mode(xxCheckConversions);
  ADD_Mode(xxInmosExtensions);
  ADD_Mode(xxFasterGOTO);
  if (Compiler_is_LittleEndian) ADD_Mode(xxLittleEndian);
  /*ADD_Mode(xxEmulate);*/
  
  #ifdef SUPPORT_INT64
  ADD_Mode(xxSupportINT64);
  #endif
  
  DefaultModes = GET_Mode();
  /*}}}*/
  /*{{{  check arguments*/
  {
    int i;
    for(i=1;i<argc;i++)
    {
      if (argv[i][0]=='-')
      {
        int j=1;
        while (argv[i][j]!='\0')
          switch (argv[i][j])
          {
            case 'C': j++; break;  /* compatibility */
            case 'd': SET_Mode(xxCDirectives,!IS_Flag(xxCDirectives,DefaultModes)); j++; break;
            case 'e':
                      /*{{{  error mode*/
                      j++;
                      switch (argv[i][j])
                      {
                        case 's' : ErrorMode=xxStoppOnError; j++; break;
                        case 't' : ErrorMode=xxTerminateOnError; j++; break;
                        case 'c' : ErrorMode=xxContinueOnError; j++; break;
                        default :  MessageI("Unknown -e Option",xxWarning,NoPosition,xxCharacter,&argv[i][j]); j++;
                      }
                      break;
                      /*}}}*/
            case 'g': SET_Mode(xxOccamDebug,!IS_Flag(xxOccamDebug,DefaultModes)); j++; break;
            case 'h': case 'H' : case 'V' :
                       /*{{{  Print help messages and options*/
                       fprintf(stderr,"SPOC : Southampton's Portable Occam Compiler, Version %s, %s\n",VersionName,VersionDate);
                       fprintf(stderr,"M. Debbage, M. Hill, S. Wykes  University Of Southampton, ESPRIT GPMIMD P5404\n\n");
                       if (argv[i][j]=='V') exit(0);
                       fprintf(stderr,"Usage: %s [-options] [filename]\n",argv[0]);
                       fprintf(stderr,"  -d : (%c) %s C Directives\n",BoolChr(IS_Mode(xxCDirectives)),DEFAULT_STR(xxCDirectives));
                       fprintf(stderr,"  -e :     select Error mode, suffix by one of:\n");
                       fprintf(stderr,"           c : (%c) Continue on error\n",BoolChr(ErrorMode==xxContinueOnError));
                       fprintf(stderr,"           s : (%c) Stop process on error\n",BoolChr(ErrorMode==xxStoppOnError));
                       fprintf(stderr,"           t : (%c) Terminate application on error\n",BoolChr(ErrorMode==xxTerminateOnError));
                       fprintf(stderr,"  -g : (%c) support occam debuGging \n",BoolChr(IS_Mode(xxOccamDebug)));
                       fprintf(stderr,"  -i : (%c) include Intrinsics\n",BoolChr(dointrinsics));
                       fprintf(stderr,"  -l : (%c) generate Library unit\n",BoolChr(IS_Mode(xxLibrary)));
                       fprintf(stderr,"  -m :     change code generation Mode, suffix by any combination of:\n");
                         /*{{{  -m modes*/
                           fprintf(stderr,"           a : (%c) %s Array checks\n",BoolChr(IS_Mode(xxCheckArrays)),DEFAULT_STR(xxCheckArrays));
                           fprintf(stderr,"           c : (%c) %s Conversion checks\n",BoolChr(IS_Mode(xxCheckConversions)),DEFAULT_STR(xxCheckConversions));
                         
                         if (IS_Flag(xxLittleEndian,DefaultModes))
                           fprintf(stderr,"           e : (%c) select big-Endian source code\n",BoolChr(!IS_Mode(xxLittleEndian)));
                         else
                           fprintf(stderr,"           e : (%c) select little-Endian source code\n",BoolChr(IS_Mode(xxLittleEndian)));
                         
                           fprintf(stderr,"           f : (%c) generate go Faster gotos\n",BoolChr(IS_Mode(xxFasterGOTO)));
                           fprintf(stderr,"           i : (%c) generate target Independant code\n",BoolChr(IS_Mode(xxEmulate)));
                         #ifdef SUPPORT_INT64     
                           fprintf(stderr,"           l : (%c) %s long-long INT64\n",BoolChr(IS_Mode(xxSupportINT64)),DEFAULT_STR(xxSupportINT64));
                         #endif                   
                           fprintf(stderr,"           m : (%c) %s coMmunication checks\n",BoolChr(IS_Mode(xxCheckInternals)),DEFAULT_STR(xxCheckInternals));
                           fprintf(stderr,"           n : (%c) generate occam line Numbers\n",BoolChr(IS_Mode(xxOccamLines)));
                           fprintf(stderr,"           r : (%c) %s Range checks\n",BoolChr(IS_Mode(xxCheckRanges)),DEFAULT_STR(xxCheckRanges));
                           fprintf(stderr,"           w : (%c) %s 16-bit Word INT\n",BoolChr(IS_Mode(xx16bitINT)),DEFAULT_STR(xx16bitINT));
                         /*}}}*/
                       fprintf(stderr,"  -o <file> : specify new output filename (use - for stdout)\n");
                       if (argv[i][j]=='H')
                       {
                         fprintf(stderr,"  -p :     control internal execution Phases, suffix by any combination of: \n");
                         /*{{{  */
                         fprintf(stderr,"           t : disable Type checking calculation phase\n");
                         fprintf(stderr,"           p : disable Parallel usage calculation phase\n");
                         fprintf(stderr,"           s : disable code Simplification/transformation phase\n");
                         fprintf(stderr,"           a : disable Attribute calculation phase\n");
                         fprintf(stderr,"           c : disable C code generation phase\n");
                         fprintf(stderr,"           o : enable Occam code generation phase (disabled C generation)\n");
                         
                         /*}}}*/
                       }
                       else
                         fprintf(stderr,"  -p :     control internal execution Phases (use option -H for details)\n");
                       fprintf(stderr,"  -s : (%c) enable simulated Server\n",BoolChr(IS_Mode(xxSimulatedServer)));
                       fprintf(stderr,"  -t#: (%c:%d) Allows tabs (of size #) in source code (no size turns off source tabs)\n",BoolChr(!no_tabs),tab_size);
                       fprintf(stderr,"  -u : (%c) Usage warnings instead of errors\n",BoolChr(IS_Mode(xxUsageWarnings)));
                       fprintf(stderr,"  -v : (%c) Verbose compilation information\n",BoolChr(lexonly));
                       fprintf(stderr,"  -w : (%c) suppress Warnings\n",BoolChr(nowarnings));
                       fprintf(stderr,"  -x :     select language eXtensions, suffix by any combination of:\n");
                       fprintf(stderr,"           i : (%c) standard Inmos extensions\n",BoolChr(IS_Mode(xxInmosExtensions)));
                       fprintf(stderr,"           u : (%c) User data type extensions\n",BoolChr(IS_Mode(xxUserDataTypes)));
                       if (argv[i][j]=='H')
                       {
                         fprintf(stderr,"  -z :     Internal debugging options, suffix by any combination of:\n");
                         /*{{{  */
                         fprintf(stderr,"           d : (%c) Debug parser\n",BoolChr(yydebug));
                         fprintf(stderr,"           e : (%c) sort Error messages\n",BoolChr(dostore));
                         fprintf(stderr,"           l : (%c) Lex only\n",BoolChr(lexonly));
                         fprintf(stderr,"           n : (%c) show symbol Node details\n",BoolChr(showdecs));
                         fprintf(stderr,"           q : (%c) Query syntax tree\n",BoolChr(doquery));
                         fprintf(stderr,"           s : (%c) show Symbol table\n",BoolChr(showsyms));
                         fprintf(stderr,"           t : (%c) show abstract syntax Tree\n",BoolChr(showtree));
                         fprintf(stderr,"           x : (%c) leXical debug\n\n",BoolChr(ldebug));
                         /*}}}*/
                       }
                       else
                         fprintf(stderr,"  -z :     enable/disable internal debugging options (use option -H for details)\n\n");
                       exit(0);
                       /*}}}*/
            case 'i': dointrinsics =!d_dointrinsics; j++; break;
            case 'l': SET_Mode(xxLibrary,!IS_Flag(xxLibrary,DefaultModes)); j++; break;
            case 'm':
                      /*{{{  code generation mode*/
                        j++;
                        while (argv[i][j]!='\0')
                        {
                        switch (argv[i][j])
                        {
                          case 'a' : SET_Mode(xxCheckArrays,!IS_Flag(xxCheckArrays,DefaultModes)); j++; break;
                          case 'c' : SET_Mode(xxCheckConversions,!IS_Flag(xxCheckConversions,DefaultModes)); j++; break;
                          case 'e' : SET_Mode(xxLittleEndian,!IS_Flag(xxLittleEndian,DefaultModes)); j++; break;
                          case 'f' : SET_Mode(xxFasterGOTO,!IS_Flag(xxFasterGOTO,DefaultModes)); j++; break;
                          case 'i' : SET_Mode(xxEmulate,!IS_Flag(xxEmulate,DefaultModes)); j++; break;
                      #ifdef SUPPORT_INT64  
                          case 'l' : SET_Mode(xxSupportINT64,!IS_Flag(xxSupportINT64,DefaultModes)); j++; break;
                      #endif                
                          case 'm' : SET_Mode(xxCheckInternals,!IS_Flag(xxCheckInternals,DefaultModes)); j++; break;
                          case 'n' : SET_Mode(xxOccamLines,!IS_Flag(xxOccamLines,DefaultModes)); j++; break;
                          case 'r' : SET_Mode(xxCheckRanges,!IS_Flag(xxCheckRanges,DefaultModes)); j++; break;
                          case 'w' : SET_Mode(xx16bitINT,!IS_Flag(xx16bitINT,DefaultModes)); j++; break;
                          default :  MessageI("Unknown -m Option",xxWarning,NoPosition,xxCharacter,&argv[i][j]); j++;
                        }
                      }
                      break;
                      /*}}}*/
            case 'o':
                      /*{{{  output file name*/
                      if (strlen(outfile) == 0 && i < argc-1)
                      {
                        outfile = argv[++i]; j = strlen(argv[i]);
                      }
                      else
                      {
                        Message("Invalid -o option",xxError,NoPosition);
                      }
                      break;
                      /*}}}*/
            case 'p':
                      /*{{{  compiler phase control*/
                      j++;
                      while (argv[i][j]!='\0')
                      {
                        switch (argv[i][j])
                        {
                          case 'a': docout_ag = false; docout = false; j++; break;
                          case 'c': docout = false; j++; break;
                          case 'o': dooccout = true; docout = false; j++; break;
                          case 'p': doparusage_ag = false; j++; break;
                          case 's': dotransform = false; docout = false; j++; break;
                          case 't': dotypechk_ag = false; docout = false; j++; break;
                          default :  MessageI("Unknown -p Option",xxWarning,NoPosition,xxCharacter,&argv[i][j]); j++;
                        }
                      }
                      break;
                      /*}}}*/
            case 's': SET_Mode(xxSimulatedServer,!IS_Flag(xxSimulatedServer,DefaultModes)); j++; break;
            case 't':
                      /*{{{  tabs*/
                      j++;
                      {
                        char c=argv[i][j];
                        if (isdigit(c))
                        {
                          no_tabs=false;
                          tab_size=c-'0';
                          c=argv[i][++j];
                          while (isdigit(c))
                          {
                            tab_size = (c*10)+(c-'0');
                            c=argv[i][++j];
                          }
                        }
                        else
                          no_tabs=true;
                      };
                      break;
                      /*}}}*/
            case 'u': SET_Mode(xxUsageWarnings,!IS_Flag(xxUsageWarnings,DefaultModes)); j++; break;
            case 'v': showinfo = !d_showinfo; j++; break;
            case 'w': nowarnings = !d_nowarnings; j++; break;
            case 'x':
                      /*{{{  select extension modes*/
                      j++;
                      while (argv[i][j]!='\0')
                      {
                        switch (argv[i][j])
                        {
                          case 'i' : SET_Mode(xxInmosExtensions,!IS_Flag(xxInmosExtensions,DefaultModes)); j++; break;
                          case 'u' : SET_Mode(xxUserDataTypes,!IS_Flag(xxUserDataTypes,DefaultModes)); j++; break;
                        }
                      }
                      break;
                      /*}}}*/
            case 'z':
                      /*{{{  internal debugging stuff*/
                      j++;
                      while (argv[i][j]!='\0')
                      {
                         switch (argv[i][j])
                         {
                            case 'd' : yydebug = !d_yydebug; j++; break;
                            case 'e' : dostore = !d_dostore; j++; break;
                            case 'l' : lexonly = !d_lexonly; j++; break;
                            case 'n' : showdecs = !d_showdecs; j++; break;
                            case 'q' : doquery = !d_doquery; j++; break;
                            case 's' : showsyms = !d_showsyms; j++; break;
                            case 't' : showtree = !d_showtree; j++; break;
                            case 'x' : ldebug = !d_ldebug; j++; break;
                            default :  MessageI("Unknown -z Option",xxWarning,NoPosition,xxCharacter,&argv[i][j]); j++;
                         }
                      }
                      break;
                      /*}}}*/
            case '-': j++; break;
            default : MessageI("Unknown Option : ",xxWarning,NoPosition,xxCharacter,&argv[i][j]); j++;
          };
      }
      else if (infile==stdin)
      {
        filename = argv[i];
        if ((infile = fopen(filename,"r"))==NULL)
        {
          Attribute.Position.File = PutString(filename,strlen(filename));
          Message("File does not exist",xxFatal,Attribute.Position);
          exit(-1);
        }
      }
    }
  }
  /*}}}*/

  if (IS_Mode(xxOccamDebug)) ADD_Mode(xxOccamLines);  
  if (dostore) StoreMessages(true);

  /*{{{  set up DEFAULT search path*/
  if (ospoc == NULL)
    Message("Could not get SPOC environment variable",xxError,NoPosition);
  
  if ((ODEFAULT_LIB = malloc (strlen(ospoc)+strlen(OLIB)+strlen(OARCH)+4)) == NULL)
    Message("Out of heap",xxFatal,NoPosition);
  else
    {
      strcpy (ODEFAULT_LIB,ospoc);
      strcat (ODEFAULT_LIB,SLASH);
      strcat (ODEFAULT_LIB,OLIB);
      strcat (ODEFAULT_LIB,SLASH);
      strcat (ODEFAULT_LIB,OARCH);
      strcat (ODEFAULT_LIB,SLASH);
    }
  /*}}}*/

  /*{{{  call init for gmd modules*/
  InitScanner(filename,infile);
  InitStringMemory();
  InitIdents();
  InitSymbols();
  /*}}}*/
  /*{{{  include intrinsic libraries*/
  /* When the simulated server is used, the back-end assumes that
     its prototype can be found at the beginning of the symbol table.
     The following two lines are reversed to set up the includes
     stack appropriately.
  */
     
  if (dointrinsics)
    if (!EnterInclude(PutString(IntrinsicsFile,strlen(IntrinsicsFile)),ExtnModes|BIT_Mode(xxPrototype)))
      MessageI("Could not open intrinsics file ", xxError, Attribute.Position, xxString, IntrinsicsFile);
  
  if (IS_Mode(xxSimulatedServer) && !IS_Mode(xxLibrary))
    if (!EnterInclude(PutString(SpServerFile,strlen(SpServerFile)),ExtnModes|BIT_Mode(xxPrototype)))
      MessageI("Could not open spserver file ", xxError, Attribute.Position, xxString, SpServerFile);
  
  /*}}}*/

  if (lexonly)
  {
    while(GetToken() != 0);
  }
  else
  {
    if (showinfo) fprintf(stderr,"Parsing code\n") ;
    (void) Parser	();
    CloseParser	();
    OccSymSize=SymSize;
    ApplyAttributeGrammars();

    /*{{{  if (showtree)*/
    if (showtree)
    {
      if (showinfo) fprintf(stderr,"Writing tree\n") ;
      WriteTree(stdout,TreeRoot);
    }
    /*}}}*/
    /*{{{  if (showsyms)*/
    if (showsyms)
    {
      if (showinfo) fprintf(stderr,"Writing symbols\n") ;
      WriteSymTable() ;
    }
    /*}}}*/
    /*{{{  if (showdecs)*/
    if (showdecs)
    {
      int i;
      for (i=1;i<=SymSize;i++)
      {
        printf("ENTRY %d scope %d ",i,SymTable[i].Scope);WriteIdent(stdout,SymTable[i].Id);
        printf("\n");
        WriteTreeNode(stdout,SymTable[i].Kind);
      }
    }
    /*}}}*/

    if (doquery) QueryTree(TreeRoot);

    /*{{{  if (dooccout)*/
    if (dooccout)
    {
      if (showinfo) fprintf(stderr,"Generating occam output\n") ;
      OutputOccam(strlen(outfile)==0 ? "-" : outfile,TreeRoot);
    }
    /*}}}*/
    /*{{{  if (ErrorCount==0 && docout)*/
    if (ErrorCount==0 && docout)
    {
      if (strlen(outfile) == 0)
      {
        char *dot = strrchr(filename,'.');
    
        if (dot != NULL) *dot = 0;
        outfile = malloc(strlen(filename) + 2);
        strcpy(outfile,filename); strcat(outfile,".c");
      }
      
      if (showinfo) fprintf(stderr,"Generating C output\n") ;
      
      OutputC(outfile,TreeRoot);
      if (showsyms) WriteSymTable() ;
    }
    /*}}}*/

    if (dostore) WriteMessages(stderr);
    if (ErrorCount>0) fprintf(stderr,"%d Errors encountered.\n",ErrorCount);
  }

  if (showinfo) fprintf(stderr,"Terminating\n") ;
 
  if (infile!=stdin) fclose(infile);

  return (ErrorCount>0)?COMPILATION_FAILED:0;
}
